<?php

/* vim: set expandtab shiftwidth=4 tabstop=4 softtabstop=4 foldmethod=marker textwidth=80: */

/**
 * Linked list structure
 * 
 * This package implements a doubly linked list structure. Each node
 * (Structure_Linked_List_Node object) in the list (Structure_Linked_List)
 * knows the previous node and the next node in the list. Unlike an array, you
 * can insert or delete nodes at arbitrary points in the list.
 *
 * Structure_Linked_List implements the Iterator interface so control structures
 * like foreach($list as $node) and while($list->next()) work as expected.
 *
 * To use this package, derive a child class from Structure_Linked_List_Node and
 * add data to the object. Then use the Structure_Linked_List class to access
 * the nodes.
 *
 * <code>
 * <?php
 * 
 * require 'Structure_Linked_List.php';
 * 
 * class Tester extends Structure_Linked_List_Link {
 *     protected $_my_number;
 * 
 *     function __construct($tester, $num) {
 *         $this->_my_number = $num;
 *         parent::__construct($tester);
 *     }
 * 
 *     function getNumb() {
 *         return $this->_my_number;
 *     }
 * 
 *     function setNumb($numb) {
 *         $this->_my_number = $numb;
 *     }
 * }
 * $tester = new Tester(null, 1);
 * $xyy = new Structure_Linked_List($tester);
 * $tester2 = new Tester($tester, 2);
 * $tester3 = new Tester($tester2, 3);
 * $tester4 = new Tester($tester, 4);
 * 
 * $link = $xyy->current();
 * print $link->getNumb();
 * print "\n";
 * 
 * // iterate through the list with while()
 * while ($link = $xyy->next()) {
 *     print $link->getNumb();
 * }
 * 
 * // rewind the list pointer to the root node
 * $link = $xyy->rewind();
 * 
 * // test foreach() iteration
 * foreach ($xyy as $bull) {
 *   print $bull->getNumb();
 * }
 * ?>
 * </code>
 * 
 * PHP version 5
 *
 * LICENSE: This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category   Structures
 * @package    Structure_Linked_List
 * @author     Dan Scott <dscott@laurentian.ca>
 * @copyright  2006 Dan Scott
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @version    CVS: $Id:$
 * @link       http://pear.php.net/package/Structure_Linked_List
 *
 * @todo Add some actual error conditions
 **/

require_once 'PEAR/ErrorStack.php';

// {{{ constants
/**
 * Add link to the end of the linked list
 */
define("STRUCTURE_LINKED_LIST_ADD_APPEND", 0);

/**
 * Add link before the target link in the linked list
 */
define("STRUCTURE_LINKED_LIST_ADD_BEFORE", 1);

/**
 * Add link after the target link in the linked list
 */
define("STRUCTURE_LINKED_LIST_ADD_AFTER", 2);

// }}}

//{{{ GLOBALS
/**
 * Error message stack
 * @global ErrorStack $GLOBALS["_Structure_Linked_List_stack"]
 */
$GLOBALS["_Structure_Linked_List_stack"] =& PEAR_ErrorStack::singleton('Structure_Linked_List');

/**
 * Error messages
 * @global array $GLOBALS["_Structure_Linked_List_messages"]
 */
$GLOBALS["_Structure_Linked_List_messages"] = array(
    STRUCTURE_LINKED_LIST_EMPTY => 'No nodes in this linked list', 
);
// }}}

//{{{ Error handling with PEAR::ErrorStack
$GLOBALS["_Structure_Linked_List_stack"]->setErrorMessageTemplate($GLOBALS["_Structure_Linked_List_messages"]);
//}}}

// {{{ class Structure_Linked_List
/**
 * The Structure_Linked_List class represents a linked list structure composed of {@link Structure_Linked_List_Node} objects.
 *
 * @category   Structures
 * @package    Structure_Linked_List
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/Structure_Linked_List
 */
class Structure_Linked_List implements Iterator {
    // {{{ properties
    /**
     * Current node in the linked list
     * @var Structure_Linked_List_Node
     */
    protected $current;

    /**
     * Root node of the linked list
     * @var Structure_Linked_List_Node
     */
    protected $root_node;

    /**
     * Error stack
     *
     * @var PEAR::ErrorStack
     */
    protected $stack;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Structure_Linked_List constructor
     *
     * @param Structure_Linked_List_Node $marc root node for the linked list
     */
    function __construct($marc=null) {
        $this->stack = new PEAR_ErrorStack('Structure_Linked_List');
        if ($marc) {
            $this->root_node = $marc;
            $this->current = $this->root_node;
        }
        else {
            $this->root_node = null;
            $this->current = null;
        }
    }
    // }}}

    // {{{ function &current()
    /**
     * Returns the current node in the linked list
     *
     * @return Structure_Linked_List_Node current node in the linked list
     */
    public function &current() {
        return $this->current;
    }
    // }}}

    // {{{ function &rewind()
    /**
     * Sets the pointer for the linked list to the root node
     *
     * @return Structure_Linked_List_Node root node in the linked list
     */
    public function &rewind() {
        if ($this->root_node) {
            $this->current = $this->root_node;
        }
        else {
            $this->current = null;
        }
        return $this->current;
    }
    // }}}

    // {{{ function &key()
    /**
     * Stub for Iterator interface that simply returns the current node
     *
     * @return Structure_Linked_List_Node current node in the linked list
     */
    public function &key() {
        return $this->current;
    }
    // }}}

    // {{{ function &valid()
    /**
     * Stub for Iterator interface that simply returns the current node
     *
     * @return Structure_Linked_List_Node current node in the linked list
     */
    public function &valid() {
        return $this->current();
    }
    // }}}

    // {{{ function &next()
    /**
     * Sets the pointer for the linked list to the next link and
     * returns that link
     *
     * @return Structure_Linked_List_Node next node in the linked list
     */
    public function &next() {
        if (!$this->current()) {
            return false;
        }
        $this->current =& $this->current()->next();
        return $this->current();
    }
    // }}}

    // {{{ function &previous()
    /**
     * Sets the pointer for the linked list to the previous link and
     * returns that link
     *
     * @return Structure_Linked_List_Node previous node in the linked list
     */
    public function &previous() {
        if (!$this->current()->previous()) {
            return false;
        }
        $this->current =& $this->current()->previous();
        return $this->current();
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $GLOBALS["_Structure_Linked_List_stack"];
    }
    // }}}

    // {{{ function addLink()
    /**
     * Adds a @see Structure_Linked_List_Node object to the linked list.
     *
     * While the @see Structure_Linked_List_Node constructor accepts a reference link
     * as the previous link in the chain, you might want to create
     * an extension of the @see Structure_Linked_List_Node object that enables you to
     * first create and manipulate the object before adding it to the
     * linked list.
     *
     * @param Structure_Linked_List_Node $newLink New link to add to the list
     * @param int $mode Specifies how the link will be inserted
     * @param Structure_Linked_List_Node $existingLink Reference link: if null, then use the final link in the list
     * @return bool Success or failure
     **/
    public function addLink($newLink, $mode=STRUCTURE_LINKED_LIST_ADD_APPEND, $existingLink=null) {
        // Try to get a valid value for the existing link
        if (!$existingLink) {
            /* We do not have a reference node for this link;
               try to use the current node in this linked list */
            $existingLink = $this->rewind();
            
            if ($existingLink) {
                /* Iterate through to the end of the list */
                while ($x = $this->next()) {
                    $existingLink = $x;
                }
            }
            else {
                /* This becomes the first link in the list */
                $this->__construct($newLink);
                return true;
            }
        }

        // Now add the link according to the requested mode
        switch ($mode) {

        case STRUCTURE_LINKED_LIST_ADD_BEFORE:
            $previousLink =& $existingLink->previous();
            if ($previousLink) {
                $previousLink->setNext($newLink);
                $newLink->setPrevious($previousLink);
            }
            else {
                /* Existing link must be root node; make new link root */
                $this->root_node =& $newLink;
                $newLink->setPrevious();
            }
            /* We do these for both cases */
            $newLink->setNext($existingLink);
            $existingLink->setPrevious($newLink);

            break;

        case STRUCTURE_LINKED_LIST_ADD_APPEND:
            if ($existingLink) {
                /* Iterate through to the end of the list */
                while ($x =& $this->next()) {
                    $existingLink =& $x;
                }
            }

            /* Fall through deliberately to STRUCTURE_LINKED_LIST_ADD_AFTER */

        case STRUCTURE_LINKED_LIST_ADD_AFTER:
            $newLink->setPrevious($existingLink);
            $newLink->setNext($existingLink->next());
            $existingLink->setNext($newLink);
            break;

        default:
            return false;
            break;

        }

        return true;
    }
    // }}}

    // {{{ function deleteLink()
    /**
     * @param {@link Structure_Linked_List_Node} $link
     */
    function deleteLink(&$link)
    {
        /* If this is the root node, and there are more nodes in the list,
         * make the next node the new root node before deleting this node.
         */
        if ($link === $this->root_node) {
            $this->root_node =& $link->next();
            $this->current =& $link->next();
        }

        /* If this is the current node, and there are other nodes in the list,
         * try making the previous node the current node so that next() works
         * as expected.
         *
         * If that fails, make the next node the current node.
         *
         * If that fails, null isn't such a bad place to be.
         */
        if ($link === $this->current) {
            if ($link->previous()) {
                $this->current =& $link->previous();
            }
            elseif ($link->next()) {
                $this->current =& $link->next();
            }
            else {
                $this->current = null;
            }
        }
        $link = null;
    }
    // }}}

}
// }}}

// {{{ class Structure_Linked_List_Node
/**
 * The Structure_Linked_List_Node class represents a node in a {@link Structure_Linked_List} linked list structure.
 *
 * @category   Structures
 * @package    Structure_Linked_List
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/Structure_Linked_List
 */
class Structure_Linked_List_Node {
    // {{{ properties
    /**
     * Next node in the linked list
     * @var Structure_Linked_List_Node
     */
    protected $next;

    /**
     * Previous node in the linked list
     * @var Structure_Linked_List_Node
     */
    protected $previous;

    /**
     * Error stack
     *
     * @var PEAR::ErrorStack
     */
    protected $stack;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Structure_Linked_List_Node constructor
     * 
     * @param Structure_Linked_List_Node $link existing {@link Structure_Linked_List_Node} to use as the previous link in the list
     */
    function __construct(&$link=null) {
        $this->stack = new PEAR_ErrorStack('Structure_Linked_List_Node');
        if ($link) {
            /* Make the passed-in link the previous link for this node */
            $this->setPrevious($link);
            /* Make the next link for the passed-in link the next link for this node instead */
            $this->setNext($link->next());
            /* Make this node the passed-in link's next link */
            $link->setNext($this);
        }
        else {
            $this->next = null;
            $this->previous = null;
        }
    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Removes node from the list, adjusting the related nodes accordingly.
     *
     * This is a problem if the node is the root node for the list.
     * At this point, however, we do not have access to the list itself. Hmm.
     */
    function __destruct() {
        $next = $this->next();
        $previous = $this->previous();
        if ($previous) {
            $previous->setNext($next);
        }
        if ($next) {
            $next->setPrevious($previous);
        }
        $this->next = null;
        $this->previous = null;
        $this->stack = null;
    }
    // }}}

    // {{{ function &next()
    /**
     * Return the next link in the linked list
     *
     * @return Structure_Linked_List_Node next node in the linked list
     */
    public function &next() {
        if ($this->next) {
            return $this->next;
        }
        else {
            return false;
        }
    }
    // }}}

    // {{{ function &previous()
    /**
     * Return the previous link in the linked list
     *
     * @return Structure_Linked_List_Node previous node in the linked list
     */
    public function &previous() {
        if ($this->previous) {
            return $this->previous;
        }
        else {
            return false;
        }
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $GLOBALS["_Structure_Linked_List_stack"];
    }
    // }}}

    // {{{ function setNext()
    /**
     * Sets the pointer for the next link in the linked list to the specified link
     *
     * @param Structure_Linked_List_Node $link new next node in the linked list
     * @return Structure_Linked_List_Node new next node in the linked list
     */
    public function setNext(&$link=null) {
        $this->next =& $link;
        return $this->next;
    }
    // }}}

    // {{{ function setPrevious()
    /**
     * Sets the pointer for the previous link in the linked list to the specified link
     *
     * @param Structure_Linked_List_Node $link new previous node in the linked list
     * @return Structure_Linked_List_Node new previous node in the linked list
     */
    public function &setPrevious(&$link=null) {
        $this->previous =& $link;
        return $this->previous;
    }
    // }}}

}
// }}}

?>
